/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 2002 The Apache Software Foundation.  All rights 
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:  
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "WSIF" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written 
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 2001, 2002, International
 * Business Machines, Inc., http://www.apache.org.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

package org.apache.wsif.providers;

import java.lang.reflect.Array;

import org.apache.wsif.WSIFException;

/**
 * A class of static utility methods for use across multiple providers
 * 
 * @author Owen Burroughs <owenb@apache.org>
 */
public class ProviderUtils {
	
	/**
	 * Convert an array of any dimensions containing only Strings of length 1, to an array
	 * of java.lang.Characters which has the same dimensions.
	 * @param obj The array of Strings
	 * @return The Character array
	 * @exception A WSIFException thrown if the conversion fails for any reason.
	 */
	public static Object stringArrayToCharacterArray(Object obj) throws WSIFException {
		return stringArrayToCharacterArray(obj, null);
	}
	
	/**
	 * Convert an array of any dimensions containing only Strings of length 1, to an array
	 * of java.lang.Characters which has the same dimensions.
	 * @param obj The array of Strings
	 * @param ret temporary array used in recursion. The first call should be passed null.
	 * @return The Character array
	 * @exception A WSIFException thrown if the conversion fails for any reason.
	 */
    protected static Object stringArrayToCharacterArray(Object obj, Object ret) throws WSIFException {
    	if (obj.getClass().isArray()) {    		
    		Object[] objs = (Object[]) obj;
    		ret = new Object[objs.length];
    		for (int i=0; i<objs.length; i++) {
    			Object temp = stringArrayToCharacterArray(objs[i], ((Object[]) ret)[i]);
    			if (i == 0) {
    				Class tempc = temp.getClass();
    				ret = Array.newInstance(tempc, objs.length);    				
    			}
    			((Object[]) ret)[i] = temp;
    		}
    		return ret;
    	} else if (obj instanceof String) {
    		String s = (String) obj;
    		if (s.length() == 1) {
    			ret = new Character(s.charAt(0));
    			return ret;
    		} else {
    			throw new WSIFException("String is longer than 1 character");
    		}
    	} else {
    		throw new WSIFException("Array entry is not a String or another array");
    	}
    }

	/**
	 * Convert an array of any dimensions containing only Strings of length 1, to an array
	 * of chars which has the same dimensions.
	 * @param obj The array of Strings
	 * @return The char array
	 * @exception A WSIFException thrown if the conversion fails for any reason.
	 */
	public static Object stringArrayToCharArray(Object obj) throws WSIFException {
		return stringArrayToCharArray(obj, null);
	}

	/**
	 * Convert an array of any dimensions containing only Strings of length 1, to an array
	 * of chars which has the same dimensions.
	 * @param obj The array of Strings
	 * @param ret temporary array used in recursion. The first call should be passed null.
	 * @return The char array
	 * @exception A WSIFException thrown if the conversion fails for any reason.
	 */	
    protected static Object stringArrayToCharArray(Object obj, Object ret) throws WSIFException {
    	if (obj.getClass().isArray()) {    		
    		Object[] objs = (Object[]) obj;
    		ret = new Object[objs.length];
    		for (int i=0; i<objs.length; i++) {
    			Object temp = stringArrayToCharArray(objs[i], ((Object[]) ret)[i]);
    			if (i == 0) {
    				Class tempc = temp.getClass();
    				ret = Array.newInstance(tempc, objs.length);    				
    			}
    			((Object[]) ret)[i] = temp;
    			if (temp instanceof Character && (i == objs.length - 1)) {
					char[] tempca = new char[objs.length];
					for (int j=0; j<objs.length; j++) {
						tempca[j] = ((Character[]) ret)[j].charValue();
					}
					ret = tempca;		    				
    			}
    		}
    		return ret;
    	} else if (obj instanceof String) {
    		String s = (String) obj;
    		if (s.length() == 1) {
    			ret = new Character(s.charAt(0));
    			return ret;
    		} else {
    			throw new WSIFException("String is longer than 1 character");
    		}
    	} else {
    		throw new WSIFException("Array entry is not a String or another array");
    	}
    }

	/**
	 * Convert an array of java.lang.Characters of any dimensions to an equivalent array 
	 * of java.lang.Strings which has the same dimensions.
	 * @param obj The array of Characters
	 * @return The array of Strings
	 * @exception A WSIFException thrown if the conversion fails for any reason.
	 */
	public static Object characterArrayToStringArray(Object obj) throws WSIFException {
		return characterArrayToStringArray(obj, null);
	}

	/**
	 * Convert an array of java.lang.Characters of any dimensions to an equivalent array 
	 * of java.lang.Strings which has the same dimensions.
	 * @param obj The array of Characters
	 * @param ret temporary array used in recursion. The first call should be passed null.
	 * @return The array of Strings
	 * @exception A WSIFException thrown if the conversion fails for any reason.
	 */
    protected static Object characterArrayToStringArray(Object obj, Object ret) throws WSIFException {
    	if (obj.getClass().isArray()) {    		
    		Object[] objs = (Object[]) obj;
    		ret = new Object[objs.length];
    		for (int i=0; i<objs.length; i++) {
    			Object temp = characterArrayToStringArray(objs[i], ((Object[]) ret)[i]);
    			if (i == 0) {
    				Class tempc = temp.getClass();
    				ret = Array.newInstance(tempc, objs.length);    				
    			}
    			((Object[]) ret)[i] = temp;
    		}
    		return ret;
    	} else if (obj instanceof Character) {
    		String s = obj.toString();
			return s;
    	} else {
    		throw new WSIFException("Array entry is not a Character or another array");
    	}
    }

	/**
	 * Convert an array of chars of any dimensions to an equivalent array 
	 * of java.lang.Strings which has the same dimensions.
	 * @param obj The array of chars
	 * @return The array of Strings
	 * @exception A WSIFException thrown if the conversion fails for any reason.
	 */
	public static Object charArrayToStringArray(Object obj) throws WSIFException {
		return charArrayToStringArray(obj, null);
	}

	/**
	 * Convert an array of chars of any dimensions to an equivalent array 
	 * of java.lang.Strings which has the same dimensions.
	 * @param obj The array of chars
	 * @param ret temporary array used in recursion. The first call should be passed null.
	 * @return The array of Strings
	 * @exception A WSIFException thrown if the conversion fails for any reason.
	 */
    protected static Object charArrayToStringArray(Object obj, Object ret) throws WSIFException {
    	if (obj.getClass().isArray()) {
    		if (obj instanceof char[]) {
    			char[] ca = (char[]) obj;
    			Character[] chra = new Character[ca.length];
    			for (int j=0; j<ca.length; j++) {
    				chra[j] = new Character(ca[j]);
    			}
    			obj = chra;
    		}    		
    		Object[] objs = (Object[]) obj;
    		ret = new Object[objs.length];
    		for (int i=0; i<objs.length; i++) {
    			Object temp = charArrayToStringArray(objs[i], ((Object[]) ret)[i]);
    			if (i == 0) {
    				Class tempc = temp.getClass();
    				ret = Array.newInstance(tempc, objs.length);    				
    			}
    			((Object[]) ret)[i] = temp;
    		}
    		return ret;
    	} else if (obj instanceof Character) {
    		String s = obj.toString();
			return s;
    	} else {
    		throw new WSIFException("Array entry is not a char or another array");
    	}
    }

	/**
	 * Convert a String to a Character. If the String is longer than one character
	 * this method will return null;
	 * @param str The String
	 * @return The Character or null if the String was longer than one character	 
	 */
    public static Character stringToCharacter(String str) {
        if (str.length() != 1)
            return null;
        return new Character(str.charAt(0));
    }

	/**
	 * Returns a default Object value for a given Class. If the Class is a primitive type
	 * the method will return the default value for that primitive type wrapped up in its
	 * object form. For example, invoking the method with int.class will return a new 
	 * java.lang.Integer with an int value of 0. If the Class does not represent a
	 * primitive type then null will be returned.
	 * @param cls The Class
	 * @return The default object value
	 */
	public static Object getDefaultObject(Class cls) {
		if (cls == null) {
			return null;
		} else if (cls.isPrimitive()) {
			if (cls.getName().equals("int")) {
				return new Integer(0);
			} else if (cls.getName().equals("char")) {
				return new Character('0');
			} else if (cls.getName().equals("long")) {
				return new Long(0);
			} else if (cls.getName().equals("short")) {
				short s = 0;
				return new Short(s);
			} else if (cls.getName().equals("double")) {
				return new Double(0);
			} else if (cls.getName().equals("boolean")) {
				return new Boolean(false);
			} else if (cls.getName().equals("byte")) {
				byte b = 0;
				return new Byte(b);
			} else {
				return new Float(0);
			}
		} else {
			return null;
		}
	}   	
}
