/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 2002 The Apache Software Foundation.  All rights 
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:  
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "WSIF" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written 
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 2001, 2002, International
 * Business Machines, Inc., http://www.apache.org.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */


package org.apache.wsif.providers.jca;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.StringTokenizer;

import javax.naming.*;
import javax.resource.cci.ConnectionFactory;
import javax.wsdl.*;
import javax.wsdl.Binding;
import javax.wsdl.extensions.ExtensibilityElement;

import org.apache.wsif.util.WSIFUtils;
import org.apache.wsif.wsdl.extensions.format.TypeMapping;
import org.apache.wsif.format.WSIFFormatHandler;


/**
 * A utility class with methods simplifying the JNDI lookup of a ConnectionFactory.
 * 
 * @author Piotr Przybylski <piotrp@ca.ibm.com>
 */


public class WSIFUtils_JCA extends WSIFUtils{

	private static final long serialVersionUID = 1L;
	private static final String DOT = ".";
	private static final String lookupPrefix = "java:comp/env/";
	private static final String emptyString = "";

	/**
	 * Returns the default lookup name created from the Service namespace,
	 * Service name and Port name. The default lookup name is used if there is no explicit lookup
	 * name value specified in the Resource Adapter address extensibility element. 
	 * 
	 * @param service
	 * @param port
	 * @return String
	 */
	public static String getJNDILookupName(Service service, Port port) {
		
		String jndiName = getPackageNameFromNamespaceURI(service.getQName().getNamespaceURI());
		if(!jndiName.endsWith(DOT))
			jndiName = jndiName + DOT;
		jndiName = jndiName.replace('.', '/');
		jndiName = jndiName + WSIFUtils_JCA.getJavaNameFromXMLName(service.getQName().getLocalPart());
		jndiName = jndiName + WSIFUtils_JCA.getJavaNameFromXMLName(port.getName());
		return jndiName;
		
	}



	/**
	 * Returns a Connection Factory from the JNDI context. The order of search 
	 * is to use the component's local naming context (java:comp/env/) first, then the global naming context.
	 * If the lookup fails or the lookup result does not match <code>res_type</code>, null is returned.
	 * 
	 * @param res_ref_name
	 * @param res_type
	 * @return ConnectionFactory
	 */
	public static ConnectionFactory lookupConnectionFactory(String res_ref_name, String res_type){
		
		ConnectionFactory connectionFactory = null;
		Context ctx = null;
		try {
			// Create the initial context
			ctx = new InitialContext();
			if (ctx == null) {
				return null;
			}
			connectionFactory = (ConnectionFactory) ctx.lookup(lookupPrefix + res_ref_name);

			if (connectionFactory == null) {
				throw new NamingException();
			}

			// Verify that the retrieved factory is of type "res_type"
			try {
				if (!(Class.forName(res_type, false, Thread.currentThread().getContextClassLoader()).isInstance(connectionFactory))) {
					return null;
				}
			}
			catch (ClassNotFoundException exn5) {
				return null;
			}

			ctx.close();
		}
		catch (NamingException exn2) {
			try {
				if (ctx == null) {
					return null;
				}
				// Try lookup without lookup prefix (java:comp/env/)
				connectionFactory = (ConnectionFactory) ctx.lookup(res_ref_name);

				if (connectionFactory == null) {
					return null;
				}
				// Verify that the retrieved factory is of type "res_type"
				try {
					if (!(Class.forName(res_type, false, Thread.currentThread().getContextClassLoader()).isInstance(connectionFactory))) {
						return null;
					}
				}
				catch (ClassNotFoundException exn5) {
					return null;
				}
				ctx.close();
			}
			catch (NamingException exn3) {
				return null;
			}
		}
		return connectionFactory;	
	}

}
