/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 2002 The Apache Software Foundation.  All rights 
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:  
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "WSIF" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written 
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 2001, 2002, International
 * Business Machines, Inc., http://www.apache.org.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

package org.apache.wsif.schema;

import java.io.Serializable;
import java.util.List;
import java.util.Map;

import javax.xml.namespace.QName;

import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

import com.ibm.wsdl.util.xml.DOMUtils;

/**
 * Super class of both ComplexType and SimpleType
 * 
 * @author Owen Burroughs <owenb@apache.org>
 */
public abstract class SchemaType implements Serializable {
	
	/**
	 * Get a flag to indicate if this type is a complexType
	 * @return The boolean flag
	 */
	public boolean isComplex() {
		return false;
	}

	/**
	 * Get a flag to indicate if this type is a simpleType
	 * @return The boolean flag
	 */
	public boolean isSimple() {
		return false;
	}

	/**
	 * Get a flag to indicate if this type is an element type
	 * @return The boolean flag
	 */
	public boolean isElement() {
		return false;
	}

	/**
	 * Get a flag to indicate if this type represents an array
	 * @return The boolean flag
	 */
	public boolean isArray() {
		return false;
	}

	/**
	 * Get the type of the elements in the array represented by this type (if applicable)
	 * @return The type
	 */	
	public QName getArrayType() {
		return null;
	}

	/**
	 * Get the dimension of the array represented by this type (if applicable)
	 * @return The dimension
	 */
	public int getArrayDimension() {
		return 0;
	}

	/**
	 * Get the "name" attribute of this type
	 * @return The type's name
	 */		
	public QName getTypeName() {
		return null;
    }

	/**
	 * Get a the direct children (SimpleType or ComplexType only) for this element
	 * @return The children
	 */		
	public List getChildren() {
		return null;
    }

	/**
	 * Get a specified attribute from a given dom element
	 * @param element The dom element
	 * @param attr The name of the attribute to retrieve
	 * @return The attribute value or null is the attriute does not exist
	 */	
    protected static QName getAttributeQName(Element element, String attr) {

        if (element == null || attr == null)
            throw new IllegalArgumentException(
                "Argument to 'getAttrQName' " + "cannot be null.");

        String name = DOMUtils.getAttribute(element, attr);

        if (name == null)
            return null;

        int index = name.lastIndexOf(":");
        String prefix = null;

        if (index != -1) {
            prefix = name.substring(0, index);
            name = name.substring(index + 1);
        }
        String uri = DOMUtils.getNamespaceURIFromPrefix(element, prefix);

        return new QName(uri, name);
    }

	/**
	 * Get a specified attribute from a given dom element
	 * @param element The dom element
	 * @param attr The name of the attribute to retrieve
	 * @param tns The targetNamespace used in resolving the attribute value
	 * @return The attribute value or null is the attriute does not exist
	 */	
    protected static QName getAttributeQName(Element element, String attr, String tns) {

        if (element == null || attr == null)
            throw new IllegalArgumentException(
                "Argument to 'getAttrQName' " + "cannot be null.");

        String name = DOMUtils.getAttribute(element, attr);

        if (name == null)
            return null;

        int index = name.lastIndexOf(":");
        String prefix = null;

        if (index != -1) {
            prefix = name.substring(0, index);
            name = name.substring(index + 1);
        }
        
        String uri = null;
        if (prefix != null) {
        	uri = DOMUtils.getNamespaceURIFromPrefix(element, prefix);
        } else {
        	uri = tns;
        }

        return new QName(uri, name);
    }

	/**
	 * Get a specified attribute from a given dom element when the attribute name is a QName
	 * @param element The dom element
	 * @param attr The name of the attribute to retrieve
	 * @return The attribute value or null is the attriute does not exist
	 */	    
    protected static QName getAttributeQName(Element element, QName attr) {

        if (element == null || attr == null)
            throw new IllegalArgumentException(
                "Argument to 'getAttrQName' " + "cannot be null.");
                
		String ns = attr.getNamespaceURI();
		String lp = attr.getLocalPart();
        String name = DOMUtils.getAttributeNS(element, ns, lp);

        if (name == null)
            return null;

        int index = name.lastIndexOf(":");
        String prefix = null;

        if (index != -1) {
            prefix = name.substring(0, index);
            name = name.substring(index + 1);
        }
        String uri = DOMUtils.getNamespaceURIFromPrefix(element, prefix);

        return new QName(uri, name);

    }

	/**
	 * Get all the attributes from a given dom element
	 * @param element The dom element
	 * @param tns The targetNamespace used in resolving the attribute value
	 * @param attributes A map to populate with the attributes
	 * @return The map of QName pairs (attribute name -> attribute value) for all the element's attributes
	 */	      
    protected static void getAllAttributes(Element el, String tns, Map attributes) {
        NamedNodeMap atts = el.getAttributes();
        if (atts != null) {
            for (int a = 0; a < atts.getLength(); a++) {
                Node attribute = atts.item(a);
                String ln = attribute.getLocalName();
                String ns = attribute.getNamespaceURI();
                
                String name = "";
                if (ns != null) {
                	name = DOMUtils.getAttributeNS(el, ns, ln);
                } else {
                	name = DOMUtils.getAttribute(el, ln);
                }
                
                int index = name.lastIndexOf(":");
                String prefix = null;

                if (index != -1) {
                    prefix = name.substring(0, index);
                    name = name.substring(index + 1);
                }

                String uri = null;
                if (prefix != null || tns == null) {
                    uri = DOMUtils.getNamespaceURIFromPrefix(el, prefix);
                } else {
                    uri = tns;
                }
				attributes.put(new QName(ns, ln) ,new QName(uri, name));                
            }
        }    	
    }		
}
