/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 2002 The Apache Software Foundation.  All rights 
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:  
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "WSIF" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written 
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 2001, 2002, International
 * Business Machines, Inc., http://www.apache.org.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

package org.apache.wsif.util.jms;

import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;

import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.Queue;
import javax.jms.TextMessage;
import org.apache.soap.rpc.SOAPContext;
import org.apache.soap.transport.TransportMessage;
import org.apache.soap.util.IOUtils;
import org.apache.soap.util.net.HTTPUtils;
import org.apache.wsif.WSIFException;
import org.apache.wsif.logging.Trc;

/**
 * This class implements a JMS to HTTP bridge. That is it takes SOAP
 * messages off of a JMS queue and posts them using HTTP. The SOAP message
 * in the HTTP response is put on a JMS reply queue. This class contains a
 * main method which takes as parameters all the JMS and HTTP information
 * needed. This bridge can be cold or warm started. Cold starting wipes
 * messages off queues on startup, whereas warm starting does not.
 * 
 * @author Mark Whitlock <whitlock@apache.org>
 */
public class JMS2HTTPBridge {
    private URL httpURL = null;
    private JMS2HTTPBridgeDestination destination = null;

    private static final ArrayList outGoingHeaders =
        new ArrayList(Arrays.asList(new String[] 
           { "SOAPAction" }));

    private static final ArrayList interestingProperties =
        new ArrayList(Arrays.asList(new String[] { "JMSPriority", "JMSDeliveryMode" }));
    private boolean verbose;

    private WSIFJMSListener list = new WSIFJMSListener() {
        public void onException(JMSException arg1) {
            Trc.entry(this, arg1);
            System.out.println("Caught an exception!");
            arg1.printStackTrace();
            Trc.exit();
        }

        public void onMessage(Message message) {
            Trc.entry(this, message);
            receiveMessage(message);
            Trc.exit();
        }
    };

    public JMS2HTTPBridge(
        String initialContextFactory,
        String jndiUrl,
        String queueConnectionFactory,
        String readQueue,
        String httpUrlString,
        String startType,
        boolean verbose)
        throws Exception {
        Trc.entry(
            this,
            initialContextFactory,
            jndiUrl,
            queueConnectionFactory,
            readQueue,
            httpUrlString,
            startType,
            new Boolean(verbose));

        System.out.println(
            "Starting the JMS2HTTPBridge with"
                + "\n"
                + "Initial Context Factory = "
                + initialContextFactory
                + "\n"
                + "JNDI URL = "
                + jndiUrl
                + "\n"
                + "Queue Connection Factory  = "
                + queueConnectionFactory
                + "\n"
                + "JNDI Read Queue = "
                + readQueue
                + "\n"
                + "HTTP URL = "
                + httpUrlString
                + "\n"
                + "Start Type = "
                + startType
                + "\n"
                + "Verbose = "
                + verbose);

        destination =
            new JMS2HTTPBridgeDestination(
                new WSIFJMSFinderForJndi(
                    null,
                    initialContextFactory,
                    jndiUrl,
                    WSIFJMSFinder.STYLE_QUEUE,
                    queueConnectionFactory,
                    readQueue,
                    null),
                null,
                WSIFJMSConstants.WAIT_FOREVER,
                startType,
                verbose);

        httpURL = new URL(httpUrlString);
        this.verbose = verbose;
        Trc.exit();
    }

    public static void main(String[] args) throws Exception {
        Trc.entry(null, args);

        String usage =
            "Usage: java "
                + JMS2HTTPBridge.class.getName()
                + " [-cold|-warm] "
                + "-icf <initialContextFactory> "
                + "-jndi <jndiUrl> "
                + "-s <sampleName> "
                + "-qcf <queueConnectionFactory> "
                + "-q <readQueue> "
                + "-http <httpUrl>"
                + "-v";

        String startType = JMS2HTTPBridgeDestination.WARMSTART;
        String initialContextFactory = "com.sun.jndi.fscontext.RefFSContextFactory";
        String jndiUrl = "file:///JNDI-Directory";
        String sampleName = null;
        String queueConnectionFactory = "WSIFSampleQCF";
        String readQueue = null;
        String httpUrlString = "http://localhost:8080/soap/servlet/rpcrouter";
        boolean verbose = false;

        for (int idx = 0; idx < args.length; idx++) {
            if (!args[idx].startsWith("-"))
                throw new Exception("Bad parameter\n" + usage);

            if (args[idx].equals("-cold")) {
                startType = JMS2HTTPBridgeDestination.COLDSTART;
            } else if (args[idx].equals("-warm")) {
                startType = JMS2HTTPBridgeDestination.WARMSTART;
            } else if (args[idx].equals("-icf")) {
                idx++;
                initialContextFactory = args[idx];
            } else if (args[idx].equals("-jndi")) {
                idx++;
                jndiUrl = args[idx];
            } else if (args[idx].equals("-s")) {
                idx++;
                sampleName = args[idx];
            } else if (args[idx].equals("-qcf")) {
                idx++;
                queueConnectionFactory = args[idx];
            } else if (args[idx].equals("-q")) {
                idx++;
                readQueue = args[idx];
            } else if (args[idx].equals("-http")) {
                idx++;
                httpUrlString = args[idx];
            } else if (args[idx].equals("-v")) {
                verbose = true;
            } else
                throw new Exception("Bad parameter\n" + usage);
        }

        if (readQueue == null && sampleName != null)
            readQueue = "SoapJms" + sampleName + "Queue";

        if (startType == null
            || initialContextFactory == null
            || jndiUrl == null
            || queueConnectionFactory == null
            || readQueue == null
            || httpUrlString == null)
            throw new Exception("Missing parameter\n" + usage);

        JMS2HTTPBridge j2h =
            new JMS2HTTPBridge(
                initialContextFactory,
                jndiUrl,
                queueConnectionFactory,
                readQueue,
                httpUrlString,
                startType,
                verbose);

        j2h.listen();
        Trc.exit();
    }

    public void listen() throws WSIFException {
        Trc.entry(this);
        destination.listen(list);
        Trc.exit();
    }

    void receiveMessage(Message msg) {
        Trc.entry(this, msg);
        String payload = null;

        try {
            if (verbose)
                System.out.println("Caught a message!");

            if (msg instanceof TextMessage) {
                String body = ((TextMessage) msg).getText();
                if (body != null) {
                    if (verbose)
                        System.out.println("Message contained '" + body + "'");

                    TransportMessage tmsg =
                        new TransportMessage(body, new SOAPContext(), new Hashtable());
                    setOutGoingHeaders(msg, tmsg); 
                    tmsg.save();

                    TransportMessage response = HTTPUtils.post(getServiceURL(msg), tmsg, 30000, null, 0);
                    payload = IOUtils.getStringFromReader(response.getEnvelopeReader());
                    if (verbose)
                        System.out.println("HTTP RESPONSE IS: '" + payload + "'");
                } else {
                    System.err.println("error: message contained no body");
                    payload = "error: message contained no body";
                }
            } else {
                System.err.println("error: message was not a TextMessage as expected");
                System.err.println(msg);
                payload = "error: message was not a TextMessage as expected";
            }

        } catch (Exception e) {
            e.printStackTrace();
            payload = e.toString();
        }

        try {
            // Put the properties from the received message onto the message we are 
            // about to send. Filter out everything but those properties that we 
            // know about and are interested in, since there'll be lots of stuff in
            // props that aren't really properties at all.
            WSIFJMSProperties props = new WSIFJMSProperties(WSIFJMSProperties.OUT);
            props.getPropertiesFromMessage(msg);
            HashMap kept = new HashMap();
            Iterator it = interestingProperties.iterator();
            while (true) {
                try {
                    if (!it.hasNext())
                        break;
                    String prop = (String) it.next();
                    if (props.containsKey(prop))
                        kept.put(prop, props.get(prop));
                } catch (Exception e) {
                    System.err.println("JMS2HTTPBridge properties caught " + e);
                }
            }

            destination.setProperties(kept);
            destination.setReplyToQueue((Queue) msg.getJMSReplyTo());
            destination.send(payload, msg.getJMSMessageID());
        } catch (JMSException je) {
            je.printStackTrace();
        } catch (WSIFException we) {
            we.printStackTrace();
        }

        Trc.exit();
        return;
    }

    private void setOutGoingHeaders(Message m, TransportMessage tmsg) {
    	for (Iterator i = outGoingHeaders.iterator(); i.hasNext(); ) {
		   try {
		   	  String name = (String) i.next();
		      String value = m.getStringProperty(name);
		      if (value != null && value.length() > 0 ) {
		         tmsg.setHeader(name, "\""+value+"\"");
		      }
		   } catch (JMSException e) {
		   	  Trc.ignoredException(e);
		   }
    	}
    }

    private URL getServiceURL(Message m) {
    	URL serviceURL = null;
		try {
		   String s = m.getStringProperty( "ServiceURL" );
		   serviceURL = new URL(s);
		} catch (Exception e) {
	   	   Trc.ignoredException(e);
		   serviceURL = httpURL;
		}
		return serviceURL;
    }

}